@@ -11,53 +11,95 @@ to plot the mass spectrum and decay chains.
 The current release supports SLHA version 1, and as far as I'm aware is also
 fully compatible with SLHA2: the block structures are read and accessed
 generically. If you have any problems, please provide an example input file and
-I'll happily investigate.
+I'll happily investigate. SLHA3 is not yet supported (or standardised) but in
+recent releases the new structures will not crash the parser. Support will be
+added once the format is standardised (and in response to demand!)
 
 The plotting script provides output in PDF, EPS and PNG via LaTeX and the TikZ
 graphics package, and as LaTeX/TikZ source for direct embedding into documents or
 user-tweaking of the generated output.
 
-Users of version 1.x should note that the interface has changed a little in
-version 2.0.0 and onward, in particular in the interface of the Block objects,
-which are now more dict-like: entries can be added and accessed via the usual
-square-brackets indexing operators, including for multiple indices as is common
-for mixing matrices e.g. NMIX[1,2] as opposed to the old NMIX.entries[1][2]
-way. This does break backward compatibility but is a big improvement both for
-internal code sanity and usability. The Block interface also now supplies
-dict-like has_key(), keys(), and items() methods, as well as more specialist
-value(), set_value() and is_single_valued() methods for improved access to ALPHA
-and any other unindexed blocks.
+Users of version 2.x should note that the interface has changed a little in
+version 3.0.0 and onward: there are now generic read()/write() functions which
+can operate on filenames or file objects, and all I/O functions now
+return/accept a single Doc object rather than a tuple of blocks, decays,
+etc. dicts. This single-object interface to the SLHA document allows for more
+coherent handling of the data, as well as better robustness against future
+changes in the format and planned support for documentation comments.
+
+API example
+-----------
+
+>>> import pyslha
+>>> # generic read from file:
+>>> d = pyslha.read('spcfiles/sps1a.spc')
+>>> d
+<PySLHA Doc: 22 blocks, 35 decays, 0 xsections>
+>>> # or, to ignore blocks known to be e.g. badly formatted:
+>>> d = pyslha.read('spcfiles/sps1a.spc', ignoreblocks=['DCINFO'])
+>>> d
+<PySLHA Doc: 21 blocks, 35 decays, 0 xsections>
+>>> d.blocks
+Blocks
+  SPINFO { 1 : SOFTSUSY; 2 : 2.0.5 }
+  MODSEL { 1,1 : sugra }
+  SMINPUTS { 1 :  1.27934000e+02; 2 :  1.16637000e-05; ...
+  ...
+>>> d.blocks.has_key('MODSEL')
+True
+>>> d.blocks['MODSEL']
+MODSEL { 1,1 : sugra }
+>>> d.blocks['MODSEL'][1,1]
+'sugra'
+>>> d.blocks['MODSEL'][1,2] = 'foo'
+>>> d.blocks['MODSEL']
+MODSEL { 1,1 : sugra; 1,2 : foo }
+
+The Block interface also supplies dict-like has_key(), keys(), and items()
+methods, as well as more specialist value(), set_value() and is_single_valued()
+methods for improved access to ALPHA and any other unindexed blocks.
+
+The decay system has a similar interface to that of the generic Block for
+accessing particle decay channel information:
+
+>>> d.decays.keys()
+[23, 24, 6, 25, 35, 36, 37, 1000021, 1000006, 2000006, 1000005, 2000005, ...
+>>> d.decays[1000021]
+1000021 : mass = 6.07713704e+02 GeV : total width = 5.50675438e+00 GeV
+   1.05840237e-01 [1000005, -5]
+   1.05840237e-01 [-1000005, 5]
+   ...
+
+The DECAY blocks in SLHA files are in fact mapped into Particle objects, each of
+which contains multiple Decay objects. Check the Python documentation for Doc,
+Block, Particle and Decay, plus the pyslha module free functions for I/O.
 
 If you use PySLHA, for either model data handling or spectrum visualisation,
 please cite the paper: http://arxiv.org/abs/1305.4194
 
 TODOs:
 
-  For 2.1.4:
-   * In set_value, if first item is non-int, treat as None-indexed
+  For 3.0.x:
+   * More refinements to I/O functions, including bound write methods on Doc.
+   * Add handling of XSECTION when standardised.
+   * Use Doc to handle document-level header comments.
+   * Use _dict/DataStore to handle block and decay summary comments.
+   * In set_value, if first item is non-int, treat as None-indexed.
    * Refine value string heuristic for strings with ints in them?
-
-  For 2.2.0:
    * Add Sphinx docs.
-   * Preserve comments from read -> write (needs full-line/inline comment
-     separation). Can use separate comment dicts in Block and Decay, and
-     attach a multiline .comment attr to the returned/written dicts.
 
-  For 3.0.0:
-   * Add a new Doc type which encapsulates the blocks, decays, xsections,
-   etc. data sections and will act as a more robust single return value from
-   the read commands / argument to the write commands. Can also handle
-   document-level header comments this way and improve the __repr__ resulting
-   from an interactive read() call without a target variable (i.e. hide the
-   uninteresting OrderedDict boilerplate).
+  For 3.1.x:
+   * Preserve _inline_ comments from read -> write (needs full-line/inline
+     comment separation). Can use separate comment dicts in Block, Decay,
+     etc. and attach a multiline .comment attr to the returned/written dicts.
 
   Later/maybe:
-   * Identify HERWIG decay matrix element to use in ISAWIG
-   * Handle RPV SUSY in ISAWIG
+   * Identify HERWIG decay matrix element to use in ISAWIG.
+   * Handle RPV SUSY in ISAWIG.
 """
 
 __author__ = "Andy Buckley <andy.buckley@cern.ch>"
-__version__ = "2.1.3"
+__version__ = "3.0.1"
 
 
 
@@ -65,7 +107,7 @@ __version__ = "2.1.3"
 ## Private utility functions
 
 def _mkdict():
-    """Try to return an empty ordered dict, but fall back to normal dict if necessary"""
+    """Return an OrderedDict if possible, or a normal dict if not."""
     try:
         from collections import OrderedDict
         return OrderedDict()
@@ -76,6 +118,28 @@ def _mkdict():
         except:
             return dict()
 
+_d = _mkdict()
+class _dict(type(_d)):
+    """
+    A cosmetic wrapper on an OrderedDict if possible, or a normal dict if not.
+
+    TODO: Add (optional) comment support
+    """
+
+    def __init__(self, name=None):
+        super(_dict, self).__init__()
+        self.name = name
+
+    def __repr__(self):
+        s = ""
+        if self.name:
+            s += self.name + "\n  "
+        if self.values():
+            s += "\n  ".join(str(val) for val in self.values())
+        else:
+            s += "No entries"
+        return s
+
 def _autotype(var):
     """Automatically convert strings to numerical types if possible."""
     if type(var) is not str:
@@ -103,6 +167,29 @@ def _autotuple(a):
         return a[0]
     return tuple(a)
 
+def _read(f):
+    """Read a file's contents, autodetecting whether the arg is a file or filename,
+    and treating '-' as as indication to read from stdin."""
+    if type(f) is str:
+        if f == "-":
+            return sys.stdin.read()
+        else:
+            with open(f, "r") as ff:
+                return ff.read()
+    else:
+        return f.read()
+
+def _write(f, txt):
+    """Write to a file, autodetecting whether the arg is a file or filename,
+    and treating '-' as as indication to write to stdout."""
+    if type(f) is str:
+        if f == "-":
+            sys.stdout.write(txt)
+        else:
+            with open(f, "w") as ff:
+                ff.write(txt)
+    else:
+        f.write(txt)
 
 
 ###############################################################################
@@ -123,6 +210,37 @@ class ParseError(Exception):
         return self.msg
 
 
+###############################################################################
+## The Doc top-level container object
+
+
+class Doc(object):
+    "Top level container for everything in an SLHA record"
+    def __init__(self, blocks, decays=None, xsections=None):
+        self.blocks = blocks
+        self.decays = decays
+        self.xsections = xsections
+        self.comment = ""
+
+    def write(self, filename=None, ignorenobr=False, precision=8):
+        """
+        Convenient method for converting an SLHA Doc object to SLHA format,
+        either returned as a string or written to a file depending on whether
+        the filename variable is None.
+        """
+        if filename is None:
+            return writeSLHA(self, ignorenobr, precision)
+        else:
+            write(filename, self, ignorenobr, precision)
+
+    def __repr__(self):
+        s = "<PySLHA Doc: %d blocks, %d decays, %d xsections" % \
+            (len(self.blocks or []), len(self.decays or []), len(self.xsections or []))
+        if self.comment:
+            s += " %s" % self.comment
+        s += ">"
+        return s
+
 
 ###############################################################################
 ## The data block, decay and particle classes
@@ -158,7 +276,7 @@ class Block(object):
     """
     def __init__(self, name, q=None, entries=None):
         self.name = name
-        self.entries = _mkdict()
+        self.entries = _dict()
         if entries is not None:
             self.entries.update(entries)
         self.q = _autotype(q)
@@ -197,16 +315,21 @@ class Block(object):
         'value()' attribute may be used in that case without an argument."""
         return len(self.entries) == 1 and self.entries.keys()[0] is None
 
-    def value(self, key=None):
+    def value(self, key=None, default=1):
         """Get a value from the block with the supplied key.
 
         If no key is given, then the block must contain only one non-indexed
         value otherwise an AccessError exception will be raised.\
         """
-        if key == None and not self.is_single_valued():
+        if key is None and not self.is_single_valued():
             raise AccessError("Tried to access unique value of multi-value block")
+        if not self.has_key(key):
+            return default
         return self.entries[key]
 
+    "Alias"
+    get = value
+
     def set_value(self, *args):
         """Set a value in the block via supplied key/value arguments.
 
@@ -238,6 +361,9 @@ class Block(object):
             else:
                 self.entries[_autotuple(args[:i_first_nonint])] = _autotuple(args[i_first_nonint:])
 
+    "Alias"
+    set = set_value
+
     def has_key(self, key):
         """Does the block have the given key?"""
         return self.entries.has_key(key)
@@ -352,7 +478,7 @@ class Particle(object):
 ###############################################################################
 ## SLHA parsing and writing functions
 
-def readSLHA(spcstr, ignorenobr=False, ignorenomass=False):
+def readSLHA(spcstr, ignorenobr=False, ignorenomass=False, ignoreblocks=[]):
     """
     Read an SLHA definition from a string, returning dictionaries of blocks and
     decays.
@@ -363,8 +489,8 @@ def readSLHA(spcstr, ignorenobr=False, ignorenomass=False):
     If the ignorenomass parameter is True, parse file even if mass block is
     absent in the file (default is to raise a ParseError).
     """
-    blocks = _mkdict()
-    decays = _mkdict()
+    blocks = _dict("Blocks")
+    decays = _dict("Decays")
     #
     import re
     currentblock = None
@@ -392,12 +518,16 @@ def readSLHA(spcstr, ignorenobr=False, ignorenomass=False):
                 if not match:
                     continue
                 blockname = match.group(1)
-                qstr = match.group(2)
-                if qstr is not None:
-                    qstr = qstr[qstr.find("=")+1:].strip()
-                currentblock = blockname
-                currentdecay = None
-                blocks[blockname] = Block(blockname, q=qstr)
+                if blockname in ignoreblocks:
+                    currentblock = None
+                    currentdecay = None
+                else:
+                    qstr = match.group(2)
+                    if qstr is not None:
+                        qstr = qstr[qstr.find("=")+1:].strip()
+                    currentblock = blockname
+                    currentdecay = None
+                    blocks[blockname] = Block(blockname, q=qstr)
             ## Handle DECAY start lines
             elif line.upper().startswith("DECAY"):
                 match = re.match(r"DECAY\s+(-?\d+)\s+([\d\.E+-]+|NAN).*", line.upper())
@@ -442,8 +572,8 @@ def readSLHA(spcstr, ignorenobr=False, ignorenomass=False):
         if not ignorenomass:
             raise ParseError("No MASS block found: cannot populate particle masses")
 
-    return blocks, decays
-
+    rtn = Doc(blocks, decays)
+    return rtn
 
 
 def writeSLHABlocks(blocks, precision=8):
@@ -457,7 +587,7 @@ def writeSLHABlocks(blocks, precision=8):
         blockstr = "BLOCK %s\n" % namestr
         entrystrs = []
         for k, v in b.items():
-            entrystr = ""
+            entrystr = sep
             if type(k) == tuple:
                 entrystr += sep.join(_autostr(i, precision) for i in k)
             elif k is not None:
@@ -488,9 +618,9 @@ def writeSLHADecays(decays, ignorenobr=False, precision=8):
 
 
 
-def writeSLHA(blocks, decays, ignorenobr=False, precision=8):
+def writeSLHA(doc, ignorenobr=False, precision=8):
     """Return an SLHA definition as a string, from the supplied blocks and decays dicts."""
-    ss = [x for x in (writeSLHABlocks(blocks, precision), writeSLHADecays(decays, ignorenobr, precision)) if x]
+    ss = [x for x in (writeSLHABlocks(doc.blocks, precision), writeSLHADecays(doc.decays, ignorenobr, precision)) if x]
     return "\n\n".join(ss)
 
 
@@ -715,8 +845,8 @@ def readISAWIG(isastr, ignorenobr=False):
     branching ratio of zero.
     """
 
-    blocks = _mkdict()
-    decays = _mkdict()
+    blocks = _dict("Blocks")
+    decays = _dict("Decays")
     LINES = isastr.splitlines()
 
     def getnextvalidline():
@@ -829,10 +959,10 @@ def readISAWIG(isastr, ignorenobr=False):
 
     # TODO: Parse RPV boolean and couplings into SLHA2 blocks
 
-    return blocks, decays
+    return Doc(blocks, decays)
 
 
-def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
+def writeISAWIG(doc, ignorenobr=False, precision=8):
     """
     Return a SUSY spectrum definition in the format produced by ISAWIG for inut to HERWIG
     as a string, from the supplied SLHA blocks and decays dicts.
@@ -843,6 +973,10 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
     If the ignorenobr parameter is True, do not write decay entries with a
     branching ratio of zero.
     """
+
+    blocks = doc.blocks
+    decays = doc.decays
+
     masses = blocks["MASS"]
 
     ## Init output string
@@ -1092,78 +1226,101 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
 ###############################################################################
 ## File-level read/write functions
 
-def readSLHAFile(spcfilename, **kwargs):
+
+def read(spcfile, **kwargs):
+    """
+    Read an SLHA or ISAWIG file (or stdin).
+
+    spcfile may either be a string filename or a file object.
+    If a s string, the assumed file format is based from the
+    filename; if a file it is assumed to be SLHA format.
+
+    Other keyword parameters are passed to readSLHA/readISAWIG.
+    """
+    txt = _read(spcfile)
+    if type(spcfile) is str and spcfile.endswith(".isa"):
+        return readISAWIG(txt, **kwargs)
+    else:
+        return readSLHA(txt, **kwargs)
+
+def readSLHAFile(spcfile, **kwargs):
     """
     Read an SLHA file, returning dictionaries of blocks and decays.
 
     Other keyword parameters are passed to readSLHA.
     """
-    f = open(spcfilename, "r")
-    rtn = readSLHA(f.read(), **kwargs)
-    f.close()
-    return rtn
+    return readSLHA(_read(spcfile), **kwargs)
 
-
-def writeSLHAFile(spcfilename, blocks, decays, **kwargs):
-    """
-    Write an SLHA file from the supplied blocks and decays dicts.
-
-    Other keyword parameters are passed to writeSLHA.
-    """
-    f = open(spcfilename, "w")
-    f.write(writeSLHA(blocks, decays, **kwargs))
-    f.close()
-
-
-def readISAWIGFile(isafilename, **kwargs):
+def readISAWIGFile(isafile, **kwargs):
     """
     Read a spectrum definition from a file in the ISAWIG format, returning
     dictionaries of blocks and decays. While this is not an SLHA format, it is
     informally supported as a useful mechanism for converting ISAWIG spectra to
     SLHA.
 
-    Other keyword parameters are passed to readSLHA.
+    isafile may either be a string filename or a file object.
+
+    Other keyword parameters are passed to readISAWIG.
     """
-    f = open(isafilename, "r")
-    rtn = readISAWIG(f.read(), **kwargs)
-    f.close()
-    return rtn
+    return readISAWIG(_read(isafile), **kwargs)
 
 
-def writeISAWIGFile(isafilename, blocks, decays, **kwargs):
+def write(spcfile, doc, **kwargs):
+    """
+    Write to an SLHA or ISAWIG file (or stdout).
+
+    spcfile may either be a string filename or a file object.
+    If a s string, the assumed file format is based from the
+    filename; if a file it is assumed to be SLHA format.
+
+    Other keyword parameters are passed to writeSLHA/writeISAWIG.
+    """
+    if type(spcfile) is str and spcfile.endswith(".isa"):
+        writeISAWIG(spcfile, doc, **kwargs)
+    else:
+        writeSLHA(spcfile, doc, **kwargs)
+
+def writeSLHAFile(spcfile, doc, **kwargs):
+    """
+    Write an SLHA file from the supplied blocks and decays dicts.
+
+    Other keyword parameters are passed to writeSLHA.
+    """
+    _write(spcfile, writeSLHA(doc, **kwargs))
+
+def writeISAWIGFile(isafile, doc, **kwargs):
     """
     Write an ISAWIG file from the supplied blocks and decays dicts.
 
+    isafile may either be a string filename or a file object.
+
     Other keyword parameters are passed to writeISAWIG.
     """
-    f = open(isafilename, "w")
-    f.write(writeISAWIG(blocks, decays, **kwargs))
-    f.close()
-
+    _write(isafile, writeISAWIG(doc, **kwargs))
 
 
 ###############################################################################
 ## Main function for module testing
 
 if __name__ == "__main__":
+
     import sys
     for a in sys.argv[1:]:
-        if a.endswith(".isa"):
-            blocks, decays = readISAWIGFile(a)
-        else:
-            blocks, decays = readSLHAFile(a)
+        doc = read(a)
+        print doc
+        print
 
-        for bname, b in sorted(blocks.iteritems()):
+        for bname, b in sorted(doc.blocks.iteritems()):
             print b
             print
 
-        print blocks.keys()
+        print doc.blocks.keys()
 
-        print blocks["MASS"][25]
+        print doc.blocks["MASS"].get(25)
         print
 
-        for p in sorted(decays.values()):
+        for p in sorted(doc.decays.values()):
             print p
             print
 
-        print writeSLHA(blocks, decays, ignorenobr=True)
+        print writeSLHA(doc, ignorenobr=True)
